/* ARCADEmini assembler (unix)
 * version 1.6.1 date 2005-10-19
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <malloc.h>

#include "id_list.h"
#include "node.h"
#include "global.h"
#include "get_val.h"

//try to get value of node
static int gv_node( st_node * p_node, st_node * p_start_node );

//try to get value of identifier
static int gv_ident( st_id_list_entry * p_entry, st_node * p_start_node );

//try to get value of constant node
static int gv_node_con( st_node * p_node, st_node * p_start_node )
{
  //value is already available
  return 0; //success

  //keep compiler happy
  p_node = p_node;
  p_start_node = p_start_node;
}

//try to get value of constant identifier node
static int gv_node_con_ident( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val;

  //value is already available
  if( p_node->data.con_ident.val_avail )
    return 0; //success

  //try to get value of identifier
  ret_val = gv_ident( p_node->data.con_ident.p_id, p_start_node ); //make sure that value is there
  if( ret_val != 0 ) //no success
    return ret_val;
  ret_val = get_val_read_ident_con( p_node->data.con_ident.p_id, &val ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //write value into node
  p_node->data.con_ident.val = val;
  p_node->data.con_ident.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of unary constant operator node
static int gv_node_con_unary( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val;

  //value is already available
  if( p_node->data.con_unary.val_avail )
    return 0; //success

  //try to get value of parameter
  ret_val = gv_node( p_node->data.con_unary.p_param, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.con_unary.p_param, &val ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //execute unary operator
  switch( p_node->data.con_unary.op )
  {
    case '~': val = ~val; break;
    case '-': val = (unsigned long)-(long)val; break;
    default:
      printf( "INTERNAL ERROR: unknown unary constant operator (%c) in column %lu of line %lu\n", p_node->data.con_unary.op, p_node->create_column, p_node->create_line );
      error_cnt++;
      val = 0;
  }

  //write value into node
  p_node->data.con_unary.val = val;
  p_node->data.con_unary.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of binary constant operator node
static int gv_node_con_binary( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val1, val2, val;

  //value is already available
  if( p_node->data.con_binary.val_avail )
    return 0; //success

  //try to get value of parameter 1
  ret_val = gv_node( p_node->data.con_binary.p_param1, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.con_binary.p_param1, &val1 ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;
  //try to get value of parameter 2
  ret_val = gv_node( p_node->data.con_binary.p_param2, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.con_binary.p_param2, &val2 ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //execute binary operator
  switch( p_node->data.con_binary.op )
  {
    case '*': val = val1 * val2; break;
    case '/': val = val1 / val2; break;
    case '%': val = val1 % val2; break;
    case '+': val = val1 + val2; break;
    case '-': val = val1 - val2; break;
    case '<': val = val1 << val2; break;
    case '>': val = val1 >> val2; break;
    case '&': val = val1 & val2; break;
    case '^': val = val1 ^ val2; break;
    case '|': val = val1 | val2; break;
    default:
      printf( "INTERNAL ERROR: unknown binary constant operator (%c) in column %lu of line %lu\n", p_node->data.con_binary.op, p_node->create_column, p_node->create_line );
      error_cnt++;
      val = 0;
  }

  //write value into node
  p_node->data.con_binary.val = val;
  p_node->data.con_binary.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of register-difference constant operator node
static int gv_node_con_regdif( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val1, val2, val;

  //value is already available
  if( p_node->data.con_regdif.val_avail )
    return 0; //success

  //try to get value of parameter 1
  ret_val = gv_node( p_node->data.con_regdif.p_param1, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.con_regdif.p_param1, &val1 ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;
  //try to get value of parameter 2
  ret_val = gv_node( p_node->data.con_regdif.p_param2, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.con_regdif.p_param2, &val2 ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //get difference
  val = val1 - val2;
  if( val1 < val2 ) //warning message if result is nagative
  {
    printf( "WARNING: register difference is negative (%ld) in column %lu of line %lu\n", val, p_node->create_column, p_node->create_line );
    warning_cnt++;
  }

  //write value into node
  p_node->data.con_regdif.val = val;
  p_node->data.con_regdif.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of constant brackets node
static int gv_node_con_bra( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val;

  //value is already available
  if( p_node->data.con_bra.val_avail )
    return 0; //success

  //try to get value of parameter
  ret_val = gv_node( p_node->data.con_bra.p_param, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.con_bra.p_param, &val ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //write value into node
  p_node->data.con_bra.val = val;
  p_node->data.con_bra.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of constant sector node
static int gv_node_con_sec( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long addr_sec;
  unsigned short addr_ofs;

  //value is already available
  if( p_node->data.con_sec.val_avail )
    return 0; //success

  //try to get address of parameter
  ret_val = gv_node( p_node->data.con_sec.p_addr, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_addr( p_node->data.con_sec.p_addr, &addr_sec, &addr_ofs ); //read address
  if( ret_val != 0 ) //no success
    return ret_val;

  //write sector part of address into node
  p_node->data.con_sec.val = addr_sec;
  p_node->data.con_sec.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of constant offset node
static int gv_node_con_ofs( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long addr_sec;
  unsigned short addr_ofs;

  //value is already available
  if( p_node->data.con_ofs.val_avail )
    return 0; //success

  //try to get address of parameter
  ret_val = gv_node( p_node->data.con_ofs.p_addr, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_addr( p_node->data.con_ofs.p_addr, &addr_sec, &addr_ofs ); //read address
  if( ret_val != 0 ) //no success
    return ret_val;

  //write offset part of address into node
  p_node->data.con_ofs.val = addr_ofs;
  p_node->data.con_ofs.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of register constant node
static int gv_node_reg_con( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val;
  unsigned char reg_no;

  //register number is already available
  if( p_node->data.reg_con.reg_no_avail )
    return 0; //success

  //try to get value of parameter
  ret_val = gv_node( p_node->data.reg_con.p_con, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.reg_con.p_con, &val ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //convert value into register number
  reg_no = (unsigned char)val;
  if( val > 0xFF )
  {
    printf( "WARNING: truncating register number (%ld -> %d) in column %lu of line %lu\n", val, reg_no, p_node->create_column, p_node->create_line );
    warning_cnt++;
  }

  //write register number into node
  p_node->data.reg_con.reg_no = reg_no;
  p_node->data.reg_con.reg_no_avail = 1; //register number is now available
  return 0; //success
}

//try to get value of register identifier node
static int gv_node_reg_ident( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned char reg_no;

  //register number is already available
  if( p_node->data.reg_ident.reg_no_avail )
    return 0; //success

  //try to get value of identifier
  ret_val = gv_ident( p_node->data.reg_ident.p_id, p_start_node ); //make sure that value is there
  if( ret_val != 0 ) //no success
    return ret_val;
  ret_val = get_val_read_ident_reg( p_node->data.reg_ident.p_id, &reg_no ); //read register number
  if( ret_val != 0 ) //no success
    return ret_val;

  //write register number into node
  p_node->data.reg_ident.reg_no = reg_no;
  p_node->data.reg_ident.reg_no_avail = 1; //register number is now available
  return 0; //success
}

//try to get value of binary register operator node
static int gv_node_reg_binary( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned char reg_no;
  unsigned long val;

  //register number is already available
  if( p_node->data.reg_binary.reg_no_avail )
    return 0; //success

  //try to get value of parameter 1 (register)
  ret_val = gv_node( p_node->data.reg_binary.p_reg, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_reg( p_node->data.reg_binary.p_reg, &reg_no ); //read register number
  if( ret_val != 0 ) //no success
    return ret_val;
  //try to get value of parameter 2 (constant)
  ret_val = gv_node( p_node->data.reg_binary.p_con, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.reg_binary.p_con, &val ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //execute binary operator
  switch( p_node->data.con_binary.op )
  {
    case '+': val = (unsigned long)reg_no + val; break;
    case '-': val = (unsigned long)reg_no - val; break;
    default:
      printf( "INTERNAL ERROR: unknown binary register operator (%c) in column %lu of line %lu\n", p_node->data.reg_binary.op, p_node->create_column, p_node->create_line );
      error_cnt++;
      val = 0;
  }

  //convert value into register number
  reg_no = (unsigned char)val;
  if( val > 0xFF )
  {
    printf( "WARNING: truncating result of binary register operator (%c, %ld -> %d) in column %lu of line %lu\n", p_node->data.reg_binary.op, val, reg_no, p_node->create_column, p_node->create_line );
    warning_cnt++;
  }

  //write register number into node
  p_node->data.reg_binary.reg_no = reg_no;
  p_node->data.reg_binary.reg_no_avail = 1; //register number is now available
  return 0; //success
}

//try to get value of register brackets node
static int gv_node_reg_bra( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned char reg_no;

  //register number is already available
  if( p_node->data.reg_bra.reg_no_avail )
    return 0; //success

  //try to get value of parameter
  ret_val = gv_node( p_node->data.reg_bra.p_param, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_reg( p_node->data.reg_bra.p_param, &reg_no ); //read register number
  if( ret_val != 0 ) //no success
    return ret_val;

  //write register number into node
  p_node->data.reg_bra.reg_no = reg_no;
  p_node->data.reg_bra.reg_no_avail = 1; //register number is now available
  return 0; //success
}

//try to get value of address identifier node
static int gv_node_addr_ident( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long addr_sec;
  unsigned short addr_ofs;

  //address is already available
  if( p_node->data.addr_ident.addr_avail )
    return 0; //success

  //try to get address of identifier
  ret_val = gv_ident( p_node->data.addr_ident.p_id, p_start_node ); //make sure that value is there
  if( ret_val != 0 ) //no success
    return ret_val;
  ret_val = get_val_read_ident_addr( p_node->data.addr_ident.p_id, &addr_sec, &addr_ofs ); //read address
  if( ret_val != 0 ) //no success
    return ret_val;

  //write address into node
  p_node->data.addr_ident.addr_sec = addr_sec;
  p_node->data.addr_ident.addr_ofs = addr_ofs;
  p_node->data.addr_ident.addr_avail = 1; //address is now available
  return 0; //success
}

//try to get value of sector address node (some kind of constant node)
static int gv_node_sec_addr( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long addr_sec;
  unsigned short addr_ofs;

  //value is already available
  if( p_node->data.sec_addr.val_avail )
    return 0; //success

  //try to get address of parameter
  ret_val = gv_node( p_node->data.sec_addr.p_addr, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_addr( p_node->data.sec_addr.p_addr, &addr_sec, &addr_ofs ); //read address
  if( ret_val != 0 ) //no success
    return ret_val;

  //warning message
  printf( "WARNING: implicitly using only sector part of address in column %lu of line %lu\n", p_node->create_column, p_node->create_line ); 
  warning_cnt++;

  //write sector part of address into node
  p_node->data.sec_addr.val = addr_sec;
  p_node->data.sec_addr.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of offset address node (some kind of constant node)
static int gv_node_ofs_addr( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long addr_sec;
  unsigned short addr_ofs;

  //value is already available
  if( p_node->data.ofs_addr.val_avail )
    return 0; //success

  //try to get address of parameter
  ret_val = gv_node( p_node->data.ofs_addr.p_addr, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_addr( p_node->data.ofs_addr.p_addr, &addr_sec, &addr_ofs ); //read address
  if( ret_val != 0 ) //no success
    return ret_val;

  //warning message if sector part of address does not match sector where it was used as offset
  if( addr_sec != p_node->data.ofs_addr.addr_sec )
  {
    printf( "WARNING: implicitly using only offset part of address in other sector in column %lu of line %lu\n", p_node->create_column, p_node->create_line ); 
    warning_cnt++;
  }

  //write offset part of address into node
  p_node->data.ofs_addr.val = addr_ofs;
  p_node->data.ofs_addr.val_avail = 1; //value is now available
  return 0; //success
}

//try to get value of register node ("REGISTER")
static int gv_node_reg( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned long val;
  unsigned char reg_no;

  //register number is already available
  if( p_node->data.reg.reg_no_avail )
    return 0; //success

  //try to get value of parameter
  ret_val = gv_node( p_node->data.reg.p_con, p_start_node ); //make sure that value is there
  if( ret_val != 0 )
    return ret_val;
  ret_val = get_val_read_node_con( p_node->data.reg.p_con, &val ); //read value
  if( ret_val != 0 ) //no success
    return ret_val;

  //convert value into register number
  reg_no = (unsigned char)val;
  if( val > 0xFF )
  {
    printf( "WARNING: truncating register number (%ld -> %d) in column %lu of line %lu\n", val, reg_no, p_node->create_column, p_node->create_line );
    warning_cnt++;
  }

  //write register number into node
  p_node->data.reg.reg_no = reg_no;
  p_node->data.reg.reg_no_avail = 1; //register number is now available
  return 0; //success
}

//try to get value of next register node ("NEXT_REG")
static int gv_node_next_reg( st_node * p_node, st_node * p_start_node )
{
  int ret_val;
  unsigned char reg_no;

  //register number is already available
  if( p_node->data.next_reg.reg_no_avail )
    return 0; //success

  //no previous register
  if( p_node->data.next_reg.p_prev_reg == &node_reg_none )
  {
    //use register $0
    reg_no = 0;
  }
  //previous register is available
  else
  {
    //try to get value of previous register
    ret_val = gv_node( p_node->data.next_reg.p_prev_reg, p_start_node ); //make sure that value is there
    if( ret_val != 0 )
      return ret_val;
    ret_val = get_val_read_node_reg( p_node->data.next_reg.p_prev_reg, &reg_no ); //read register number
    if( ret_val != 0 ) //no success
      return ret_val;

    //use next register
    reg_no++;
    //error message if overflow
    if( reg_no == 0 )
    {
      printf( "ERROR: overflow of register number at \"NEXT_REG\" in column %lu of line %lu\n", p_node->create_column, p_node->create_line );
      error_cnt++;
    }
  }

  //write register number into node
  p_node->data.next_reg.reg_no = reg_no;
  p_node->data.next_reg.reg_no_avail = 1; //register number is now available
  return 0; //success
}

//try to get value of command node
static int gv_node_cmd( st_node * p_node, st_node * p_start_node )
{
  //value (the address) is already available
  return 0; //success

  //keep compiler happy
  p_node = p_node;
  p_start_node = p_start_node;
}

//try to get value of label node
static int gv_node_lab( st_node * p_node, st_node * p_start_node )
{
  //try to get value of real node (e.g. the command node)
  return gv_node( p_node->data.lab.p_real_node, p_start_node );
}

//try to get value of node
//cyclical dependency is detected if we arrive at start node
// - called with p_start_node == NULL if p_node is the first node of a value search
//retuns 0 on success, 1 on cyclical dependency , -1 on other errors
static int gv_node( st_node * p_node, st_node * p_start_node )
{
  //no node
  if( p_node == NULL )
    return -1; //error

  //we found a cyclical dependency if we have arrived at start node
  if( p_node == p_start_node )
    return 1; //cyclical dependency found

  //if no start node is supplied we are at the first node of a value search and remember this node as start node
  if( p_start_node == NULL )
    p_start_node = p_node;

  //different node types
  switch( p_node->type )
  {
    case node_generic: return -1; //generic node is an error
    case node_con: return gv_node_con( p_node, p_start_node );
    case node_con_ident: return gv_node_con_ident( p_node, p_start_node );
    case node_con_unary: return gv_node_con_unary( p_node, p_start_node );
    case node_con_binary: return gv_node_con_binary( p_node, p_start_node );
    case node_con_regdif: return gv_node_con_regdif( p_node, p_start_node );
    case node_con_bra: return gv_node_con_bra( p_node, p_start_node );
    case node_con_sec: return gv_node_con_sec( p_node, p_start_node );
    case node_con_ofs: return gv_node_con_ofs( p_node, p_start_node );
    case node_reg_con: return gv_node_reg_con( p_node, p_start_node );
    case node_reg_ident: return gv_node_reg_ident( p_node, p_start_node );
    case node_reg_binary: return gv_node_reg_binary( p_node, p_start_node );
    case node_reg_bra: return gv_node_reg_bra( p_node, p_start_node );
    case node_addr_ident: return gv_node_addr_ident( p_node, p_start_node );
    case node_sec_addr: return gv_node_sec_addr( p_node, p_start_node );
    case node_ofs_addr: return gv_node_ofs_addr( p_node, p_start_node );
    case node_reg: return gv_node_reg( p_node, p_start_node );
    case node_next_reg: return gv_node_next_reg( p_node, p_start_node );
    case node_str: return -1; //we never need the value of a string
    case node_cmd: return gv_node_cmd( p_node, p_start_node );
    case node_lab: return gv_node_lab( p_node, p_start_node );
    case node_list: return -1; //we never need the value of a list
    default: return -1; //unknown node is an error
  }
}

//try to get value of identifier
//cyclical dependency is detected if we arrive at start node
// - called with p_start_node == NULL if p_node is the first node of a value search
//retuns 0 on success, 1 on cyclical dependency , -1 on other errors
static int gv_ident( st_id_list_entry * p_entry, st_node * p_start_node )
{
  int ret_val;

  //no entry
  if( p_entry == NULL )
    return -1; //error

  //try to get value of declaration node
  if( p_entry->p_decl_node == NULL ) //no declaration node
    return -1;
  ret_val = gv_node( p_entry->p_decl_node, p_start_node );
  if( ret_val != 0 )
    return ret_val;

  //copy value into identifier list entry
  switch( p_entry->type )
  {
    case id_list_unused:
      //no data to set here
      break;
    case id_list_con:
      if( get_val_read_node_con( p_entry->p_decl_node, &p_entry->data.con.val ) != 0 ) //read value of constant
        return -1; //error
      p_entry->data.con.val_avail = 1; //value is now available
      break;
    case id_list_reg:
      if( get_val_read_node_reg( p_entry->p_decl_node, &p_entry->data.reg.reg_no ) != 0 ) //read number of register
        return -1; //error
      p_entry->data.reg.reg_no_avail = 1; //register number is now available
      break;
    case id_list_addr:
      if( get_val_read_node_addr( p_entry->p_decl_node, &p_entry->data.addr.addr_sec, &p_entry->data.addr.addr_ofs ) != 0 ) //read address from node
        return -1; //error
      p_entry->data.addr.addr_avail = 1; //address is now available
      break;
  }

  //return success
  return 0;
}

//try to get value of node
//retuns 0 on success, 1 on cyclical dependency , -1 on other errors
int get_val_node( st_node * p_node )
{
  return gv_node( p_node, NULL );
}

//try to get value of identifier
//retuns 0 on success, 1 on cyclical dependency , -1 on other errors
int get_val_ident( st_id_list_entry * p_entry )
{
  return gv_ident( p_entry, NULL );
}

//read value of constant node
//returns 0 on success, -1 on error (e.g. when value is unknown yet)
int get_val_read_node_con( st_node * p_node, unsigned long * p_val )
{
  //no node
  if( p_node == NULL )
    return -1; //error

  //different type of nodes
  switch( p_node->type )
  {
    case node_con:
      *p_val = p_node->data.con.val; //read value
      return 0; //success
    case node_con_ident:
      if( !p_node->data.con_ident.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_ident.val; //read value
      return 0; //success
    case node_con_unary:
      if( !p_node->data.con_unary.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_unary.val; //read value
      return 0; //success
    case node_con_binary:
      if( !p_node->data.con_binary.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_binary.val; //read value
      return 0; //success
    case node_con_regdif:
      if( !p_node->data.con_regdif.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_regdif.val; //read value
      return 0; //success
    case node_con_bra:
      if( !p_node->data.con_bra.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_bra.val; //read value
      return 0; //success
    case node_con_sec:
      if( !p_node->data.con_sec.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_sec.val; //read value
      return 0; //success
    case node_con_ofs:
      if( !p_node->data.con_ofs.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.con_ofs.val; //read value
      return 0; //success
    case node_sec_addr:
      if( !p_node->data.sec_addr.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.sec_addr.val; //read value
      return 0; //success
    case node_ofs_addr:
      if( !p_node->data.ofs_addr.val_avail ) //value not available
        return -1; //error
      *p_val = p_node->data.ofs_addr.val; //read value
      return 0; //success
    case node_lab:
      return get_val_read_node_con( p_node->data.lab.p_real_node, p_val );
    default: //not a constant node
      return -1; //error
  }
}

//read number of register node
//returns 0 on success, -1 on error (e.g. when register number is unknown yet)
int get_val_read_node_reg( st_node * p_node, unsigned char * p_reg_no )
{
  //no node
  if( p_node == NULL )
    return -1; //error

  //different type of nodes
  switch( p_node->type )
  {
    case node_reg_con:
      if( !p_node->data.reg_con.reg_no_avail ) //register number not available
        return -1; //error
      *p_reg_no = p_node->data.reg_con.reg_no; //read register number
      return 0; //success
    case node_reg_ident:
      if( !p_node->data.reg_ident.reg_no_avail ) //register number not available
        return -1; //error
      *p_reg_no = p_node->data.reg_ident.reg_no; //read register number
      return 0; //success
    case node_reg_binary:
      if( !p_node->data.reg_binary.reg_no_avail ) //register number not available
        return -1; //error
      *p_reg_no = p_node->data.reg_binary.reg_no; //read register number
      return 0; //success
    case node_reg_bra:
      if( !p_node->data.reg_bra.reg_no_avail ) //register number not available
        return -1; //error
      *p_reg_no = p_node->data.reg_bra.reg_no; //read register number
      return 0; //success
    case node_reg:
      if( !p_node->data.reg.reg_no_avail ) //register number not available
        return -1; //error
      *p_reg_no = p_node->data.reg.reg_no; //read register number
      return 0; //success
    case node_next_reg:
      if( !p_node->data.next_reg.reg_no_avail ) //register number not available
        return -1; //error
      *p_reg_no = p_node->data.next_reg.reg_no; //read register number
      return 0; //success
    case node_lab:
      return get_val_read_node_reg( p_node->data.lab.p_real_node, p_reg_no );
    default: //not a register node
      return -1; //error
  }
}

//read address from address node
//returns 0 on success, -1 on error (e.g. when address is unknown yet)
int get_val_read_node_addr( st_node * p_node, unsigned long * p_addr_sec, unsigned short * p_addr_ofs )
{
  //no node
  if( p_node == NULL )
    return -1; //error

  //different type of nodes
  switch( p_node->type )
  {
    case node_addr_ident:
      if( !p_node->data.addr_ident.addr_avail ) //address not available
        return -1; //error
      *p_addr_sec = p_node->data.addr_ident.addr_sec; //read address
      *p_addr_ofs = p_node->data.addr_ident.addr_ofs;
      return 0; //success
    case node_cmd:
      *p_addr_sec = p_node->data.cmd.addr_sec; //read address
      *p_addr_ofs = p_node->data.cmd.addr_ofs;
      return 0; //success
    case node_lab:
      return get_val_read_node_addr( p_node->data.lab.p_real_node, p_addr_sec, p_addr_ofs );
    default: //not an address node
      return -1; //error
  }
}

//read value of constant identifier
//returns 0 on success, -1 on error (e.g. when value is unknown yet)
int get_val_read_ident_con( st_id_list_entry * p_entry, unsigned long * p_val )
{
  //no entry
  if( p_entry == NULL )
    return -1; //error

  //not a constant identifier list entry
  if( p_entry->type != id_list_con )
    return -1; //error

  //value not available
  if( !p_entry->data.con.val_avail )
    return -1; //error

  //read value
  *p_val = p_entry->data.con.val;
  return 0; //success
}

//read number of register identifier
//returns 0 on success, -1 on error (e.g. when register number is unknown yet)
int get_val_read_ident_reg( st_id_list_entry * p_entry, unsigned char * p_reg_no )
{
  //no entry
  if( p_entry == NULL )
    return -1; //error

  //not a register identifier list entry
  if( p_entry->type != id_list_reg )
    return -1; //error

  //register number not available
  if( !p_entry->data.reg.reg_no_avail )
    return -1; //error

  //read register number
  *p_reg_no = p_entry->data.reg.reg_no;
  return 0; //success
}

//read address from address identifier
//returns 0 on success, -1 on error (e.g. when address is unknown yet)
int get_val_read_ident_addr( st_id_list_entry * p_entry, unsigned long * p_addr_sec, unsigned short * p_addr_ofs )
{
  //no entry
  if( p_entry == NULL )
    return -1; //error

  //not an address identifier list entry
  if( p_entry->type != id_list_addr )
    return -1; //error

  //address not available
  if( !p_entry->data.addr.addr_avail )
    return -1; //error

  //read address
  *p_addr_sec = p_entry->data.addr.addr_sec;
  *p_addr_ofs = p_entry->data.addr.addr_ofs;
  return 0; //success
}

//get value of command parameter
//value is stored in big endian format
//value is stored beginning at bit <first_bit> of buffer (counted from D7 to D0, i.e. D7 of first byte = bit 0)
//*p_bits_written is filled with the number of bits written
//returns 0 on success, -1 on error (e.g. when value is unknown yet)
int get_val_get_cmd_param( st_node * p_cmd, unsigned char param_no, unsigned char * p_buffer, unsigned int buffer_len, unsigned int first_bit, unsigned int * p_bits_written )
{
  st_node_cmd_param * p_cmd_param;
  int ret_val;
  unsigned long val;
  unsigned char reg_no;
  unsigned long addr_sec;
  unsigned short addr_ofs;
  unsigned long long value, data, mask;
  unsigned int value_bits, data_bits, bit;
  unsigned char buffer_mask;

  //no command node, invalid paramameter number or no buffer
  if( p_cmd == NULL || p_cmd->type != node_cmd || param_no >= p_cmd->data.cmd.param_cnt || p_buffer == NULL )
    return -1; //error

  //get pointer to parameter
  p_cmd_param = &p_cmd->data.cmd.params[param_no];

  //value of command parameter has to be taken from node
  if( p_cmd_param->p_node != NULL )
  {
    //get value (in big endian format) and length (in bits)
    switch( p_cmd_param->p_node->type )
    {
      //constant nodes
      case node_con:
      case node_con_ident:
      case node_con_unary:
      case node_con_binary:
      case node_con_regdif:
      case node_con_bra:
      case node_con_sec:
      case node_con_ofs:
      case node_sec_addr:
      case node_ofs_addr:
        //read value of constant
        ret_val = get_val_read_node_con( p_cmd_param->p_node, &val );
        if( ret_val != 0 ) //no success
          return ret_val;
        value = val;
        value_bits = 32;
        break;
      //register nodes
      case node_reg_con:
      case node_reg_ident:
      case node_reg_binary:
      case node_reg_bra:
      case node_reg:
      case node_next_reg:
        //read register number
        ret_val = get_val_read_node_reg( p_cmd_param->p_node, &reg_no );
        if( ret_val != 0 ) //no success
          return ret_val;
        value = reg_no;
        value_bits = 8;
        break;
      //address node
      case node_addr_ident:
        //read address
        ret_val = get_val_read_node_addr( p_cmd_param->p_node, &addr_sec, &addr_ofs );
        if( ret_val != 0 ) //no success
          return ret_val;
        //(address is stored with 9 bit offset in high part and 32 bit relative sector in low part)
        value = (unsigned long long)(addr_ofs & 0x01FF) << 32 | (unsigned long long)(addr_sec - p_cmd->data.cmd.addr_sec);
        value_bits = 41;
        break;
      //unexpected or unknown node type
      default:
        //return error
        return -1;
    }
  }
  //value of command parameter is fixed
  else
  {
    //get value (in big endian format) and length (in bits)
    value = p_cmd_param->val;
    value_bits = 32;
  }

  //truncate value if it is too long
  if( value_bits > p_cmd_param->bits )
  {
    mask = ((unsigned long long)1 << p_cmd_param->bits) - (unsigned long long)1;
    data = value & mask;
    data_bits = p_cmd_param->bits;
    //print warning message if truncation changed value
    //("0...0|x...x" and "1...1|1x...x" are ok)
    if( (value & ~mask) != 0 && (value | mask >> 1) != ((unsigned long long)1 << value_bits) - (unsigned long long)1 )
    {
      printf( "WARNING: truncating value of parameter %d (%lld -> %lld) of command in column %lu of line %lu\n",
        param_no, value, data, p_cmd->create_column, p_cmd->create_line ); 
      warning_cnt++;
    }
  }
  //copy value if it is short enough
  else
  {
    data = value;
    data_bits = value_bits;
  }

  //check length of buffer
  if( data_bits + first_bit > (buffer_len << 3) )
    return -1; //error

  //correct address and bit number
  //so that bit number is less than 8
  p_buffer += first_bit >> 3;
  first_bit &= 0x7;

  //write data in big endian format into buffer
  mask = (unsigned long long)1 << (data_bits - 1); //set mask to extract first bit of data
  buffer_mask = 0x80 >> first_bit; //start at bit specified by first_bit
  for( bit = 0; bit < data_bits; bit++ ) //write all bits of data
  {
    //write current data bit into buffer
    if( data & mask )
      *p_buffer |= buffer_mask;
    else
      *p_buffer &= ~buffer_mask;
    //go to next bit of data
    data <<= 1;
    //go to next bit in buffer
    buffer_mask >>= 1;
    if( buffer_mask == 0 )
    {
      buffer_mask = 0x80;
      p_buffer++;
    }
  }

  //return number of bits written
  *p_bits_written = bit;

  //return success
  return 0;
}

