/* ARCADEmini assembler (unix)
 * version 1.6.1 date 2005-10-19
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#ifndef INC_id_list
#define INC_id_list

//this tis the interface of the identifier list

//enum for types of idnetifiers
typedef enum t_id_list_type
{
  id_list_unused, //used internally for list anchor - do not use
  id_list_con,
  id_list_reg,
  id_list_addr,
} et_id_list_type;

//type of data for constant identifier
typedef struct t_id_list_con
{
  int val_avail; //flag if value is already available
  unsigned long val; //value of the constant
} st_id_list_con;

//type of data for register identifier
typedef struct t_id_list_reg
{
  int reg_no_avail; //flag if number is already available
  unsigned char reg_no; //number of the rgister
} st_id_list_reg;

//type of data for address identifier
typedef struct t_id_list_addr
{
  int addr_avail; //flag if address is already available
  unsigned long addr_sec; //sector part of the address
  unsigned short addr_ofs; //offset part of the address
} st_id_list_addr;

//type of data for an identifier
typedef union t_id_list_data
{
  char unused; //unused (used to initialize list anchor and keep compiler happy)
  struct t_id_list_con con; //data for constant identifier
  struct t_id_list_reg reg; //data for register identifier
  struct t_id_list_addr addr; //data for address identifier
} ut_id_list_data;

//type for an entry of the list
typedef struct t_id_list_entry
{
  //common members
  char * p_name; //malloc-ed identifier name
  et_id_list_type type; //type of identifier
  unsigned long firstuse_column, firstuse_line; //location of first usage (both 0 initially)
  unsigned long decl_column, decl_line; //location of declaration (both 0 initially)
  struct t_node * p_decl_node; //top node of declaration (NULL initially)
  //data for every identifier type
  union t_id_list_data data;
  //private members
  struct t_id_list_entry * _p_next, * _p_prev; //pointer to previous and next intem in doulby linked list
} st_id_list_entry;

#ifdef DEBUG
//debug function: convert identifier type number to identifier type sting
char * id_list_dbg_type2str( et_id_list_type type );
#endif

//convert identifier type to human readable identifier type string
char * id_list_type2typestr( et_id_list_type type );

//convert identifier type to prefix symbol
char * id_list_type2prefix( et_id_list_type type );

//search an identifier in the list
//returns pointer to list entry or NULL if not found
st_id_list_entry * id_list_lookup( char * p_name, et_id_list_type type );

//add an identifier to the list
//does not add an identifier of the same type a second time
//initialized location of first usage and of declaration to 0, 0 when entry is created newly
//returns pointer to list entry or NULL in case of error
st_id_list_entry * id_list_add( char * p_name, et_id_list_type type );

//enumerate identifier list
//calls a function for every entry in the identifier list
void id_list_enum( void (*enum_func)( st_id_list_entry * p_entry ) );

//free identifier list entry
void id_list_free( st_id_list_entry * p_entry );

//free entire identifier list
void id_list_free_all( );

#endif //#ifndef INC_id_list
